/*  vim: set sw=8 ts=8 si et :
* Author: Guido Socher, Copyright: GPL 2
* mini voltmeter with
* LED display.
* http://tuxgraphics.org/electronics/
*
* Chip type           : Atmega8
* Clock frequency     : Internal clock 8 Mhz 
*/
#include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>
#define F_CPU 8000000UL  // 8 MHz
#include <util/delay.h>
#include "analog.h"

// !!! there is a long explanation as to how this code with the
// !!! custom formula works further down.

//---
// Do not change any display point settings. Just hardware
// the display decimal point to be were you need it.
//
// One digit behind decimal point:
//#define DP_DIG 1
// No decimal point:
#define DP_DIG 0
// Two digits after decimal point: (do not chop any zeros):
//#define DP_DIG 2
//-------------END EDIT HERE----------------------------------------

//Connections of the seven segment elements: 
//      10 9 7 6 4 2 1 = pin
//   0| -  f a b c d e | 0111111 | 3f
//   1| -  - - b c - - | 0001100 | 0c
//   2| g  - a b - d e | 1011011 | 5b
//   3| g  - a b c d - | 1011110 | 5e
//   4| g  f - b c - - | 1101100 | 6c
//   5| g  f a - c d - | 1110110 | 76
//   6| g  f a - c d e | 1110111 | 77
//   7| -  - a b c - - | 0011100 | 1c
//   8| g  f a b c d e | 1111111 | 7f
//   9| g  f a b c d - | 1111110 | 7e
//   -| g  - - - - - - | 1000000 | 40 minus=0x40
// digit to seven segment LED mapping:
static unsigned char d2led[]={0x3f,0x0c,0x5b,0x5e,0x6c,0x76,0x77,0x1c,0x7f,0x7e};

void delay_ms(unsigned char ms)
{
        // if we would use the library function _delay_ms
        // everywhere then the code would become very fat
        // We just need integers therefore we wrap it.
        while(ms){
                _delay_ms(0.96);
                ms--;
        }
}

void init7segment(void)
{
        // least significant digit, cathode transistor:
        DDRD|= (1<<DDD7);
        // middle digit:
        DDRB|= (1<<DDB6);
        // left digit:
        DDRB|= (1<<DDB7);
        // digit pins:
        DDRD|= 0x7F;
        // off
        PORTD=0;
}

// display: --
void outofrange(void){
        PORTD=0x40; // just a dash
        PORTD|= (1<<PD7);// on
        delay_ms(4);
        PORTD&=~(1<<PD7);
        PORTB|= (1<<PB6);// on for a short time
        delay_ms(4);
        PORTB &= ~(1<<PB6);
}

// Update the sevent segement display.
// You must loop over this function all the time.
inline void upd7segment(int16_t number)
{
        uint8_t minus=0;
        if (number<0){
                number=abs(number);
                if (number>99){
                        // out of range for 3 digits
                        number=0;
                }
                minus=1;
        }
        // LS digit:
        PORTD=d2led[(number%10)];
        //
        PORTD|=(1<<PD7);// digit on for a short time:
        delay_ms(4);
        number/=10; // divide by 10 (shift off right most digit)
        PORTD&=~(1<<PD7);
#if (DP_DIG == 0)
        if (number>0){ // chop off the twp left most digits if it is zero
#endif
                // middle digit:
                PORTD=d2led[(number%10)];
                //
                PORTB|= (1<<PB6);// digit on for a short time:
                delay_ms(4);
                number/=10; // divide by 10 (shift off right most digit)
                PORTB &= ~(1<<PB6);
                // minus on the left
                if (minus){
                        PORTD=0x40;
                        PORTB|= (1<<PB7);// on
                        delay_ms(4);
                        PORTB &= ~(1<<PB7);
                        // go out of this function
                        return;
                }
#if (DP_DIG == 1 || DP_DIG == 0)
                if (number>0){ // chop off left most digit if it is zero
#endif
                        //
                        // left digit, we assume there that the number is
                        // smaller than 999, no modulo here:
                        PORTD=d2led[number];
                        //
                        PORTB|= (1<<PB7);// digit on for a short time:
                        delay_ms(4);
                        PORTB &= ~(1<<PB7);
#if (DP_DIG == 1 || DP_DIG == 0)
                }
#endif
#if (DP_DIG == 0)
        }else{
                if (minus){
                        PORTD=0x40;
                        PORTB|= (1<<PB6);// on
                        delay_ms(4);
                        PORTB &= ~(1<<PB6);
                }
        }
#endif
}

/*  This an example of a voltage read from a temperature sensor 
 *  in a car. The voltages from the sensor are then converted 
 *  into some custom display value according to a formula.
 *
 *  At a voltage of 0.73V it shall display -58
 *  and at a volage of 2.32 it shall display 347
 *  From those 2 points we calculate the formula.
 *
 *  The formula is:  display=254.72*Input-243.95
 *
 *  We add a voltage divider with 47K=Ry and 100K=Rx 
 *  In other words we divide by 1.47. An external voltage of 3.76V would
 *  correspond to an ADC reading of 1023 since the internal refernce voltage
 *  in the microcontroller is 2.56.
 *  2.56*1.47=3.76
 *
 *  In other words our formula becomes:
 *
 *  display=254.72*(ADC/1023 *3.76) -243.95
 *  or
 *  display=ADC*0.9362 -243.95
 *  or
 *  display=ADC/1.0681 -243.95
 *
 * Now compile this load it and calibrate it if needed by changing the 
 * formula slightly.  Especially the slope can be different due to tolerances
 * in the voltage divider.
 *
 */
int main(void)
{
        unsigned char lc=0; // loop counter
        float calc=0;
        unsigned int adc=0;
        init7segment();

        while (1) {
                lc++;
                // 
                if (lc==1){
                        // once in a while:
                        adc=convertanalog(0);
                        calc=(adc/1.046) - 243.0;
                }
                // step loop counter up to 0x1f (=00011111, we can use binary AND):
                lc=lc& 0x1f;
                if (adc < 1020){
                        // off if out of range, you don't have to do
                        // this but the customer wanted it
                        upd7segment((uint16_t)calc);
                }else{
                        outofrange();
                }
        }
        return(0);
}


